import os
import subprocess
import multiprocessing
import json
import numpy as np
import cv2
import random

from scipy.misc import imread

num_gpus = 1

#COD CIFS from materials science journals without H (Richard specified no H)
CIFS_DIR = r"\\flexo.ads.warwick.ac.uk\shared41\Microscopy\Jeffrey-Ede\crystal_structures\standardized_inorganic_no_H"

PARENT_DIR = r"\\flexo.ads.warwick.ac.uk\shared41\Microscopy\Jeffrey-Ede\models\wavefunctions"
default_json_filepath = os.path.join(PARENT_DIR, "default.json")

EXE_DIR = r"\\flexo.ads.warwick.ac.uk\shared41\Microscopy\Jeffrey-Ede\models\wavefunctions\clTEM_files"
exe_filepath = os.path.join(EXE_DIR, "clTEM_cmd.exe")

OUTPUT_DIR = os.path.join(PARENT_DIR, "run_specific_output")
if not os.path.exists(OUTPUT_DIR):
    os.makedirs(OUTPUT_DIR)

NUM_REPEATS = 1

CONFIG_DIR = os.path.join(PARENT_DIR, "run_specific_temp")
config_filepath = os.path.join(CONFIG_DIR, "current_config.json")
if not os.path.exists(CONFIG_DIR):
    os.makedirs(CONFIG_DIR)

with open(default_json_filepath, "r") as f:
    default_config = json.load(f)

# print(default_config)

def random_config():
    """Change default configuration to random configuration."""

    config = default_config.copy()
    
    # Things to randomise
    # Voltage (use some presets)
    # aperture size
    # convergence
    # defocus spread
    voltages = [300, 200, 80]

    config["microscope"]["voltage"] = 80#random.choice(voltages)
    config["microscope"]["aperture"] = 15#np.random.uniform(5, 30)
    config["microscope"]["delta"] = 10#np.random.uniform(0, 20)
    config["microscope"]["alpha"] = 1#np.random.uniform(0.1, 2)

    # aberrations
    config["microscope"]["aberrations"]["C10"]["val"] = np.random.uniform(-30, 30)

    config["microscope"]["aberrations"]["C12"]["mag"] = np.random.uniform(-50, 50)
    config["microscope"]["aberrations"]["C12"]["ang"] = np.random.uniform(0, 180)

    config["microscope"]["aberrations"]["C21"]["mag"] = np.random.uniform(-1000, 1000)
    config["microscope"]["aberrations"]["C21"]["ang"] = np.random.uniform(0, 180)

    config["microscope"]["aberrations"]["C23"]["mag"] = np.random.uniform(-1000, 1000)
    config["microscope"]["aberrations"]["C23"]["ang"] = np.random.uniform(0, 180)

    config["microscope"]["aberrations"]["C30"]["val"] = np.random.uniform(-500, 500)

    return config

def do_sim(cif_filepath):
    #
    # This is a real bodge to match the device to the thread....
    #
    device = 1 #int(multiprocessing.current_process().name[-1]) - 1

    device_string = "0:%s" % device

    print("Simulating on device:" + device_string + " using file: " + cif_filepath)
    #print("\n\n\n")

    for repetition in range(NUM_REPEATS):  # Number of times to go through CIFs
        #
        # Create output folder
        #

        # make a folder for each cif
        cif_name = "0"
        out_filepath = os.path.join(OUTPUT_DIR, cif_name)
        
        # make a folder for each repetition
        out_repeat_filepath = os.path.join(out_filepath, str(repetition))

        if not os.path.exists(out_repeat_filepath):
            os.makedirs(out_repeat_filepath)

        #
        # Randomise the simulation parameters
        #

        # Save random configuration
        config = random_config()
        with open(config_filepath, "w") as f:
            json.dump(config, f)

        #
        # Randomise the structure inputs
        #

        # randomise the cell depth (between 5 nm and 100 nm)
        cell_depth = 100#np.random.uniform(50, 1000)
        cell_widths = 75#np.random.uniform(50, 100)
        cell_string = "%s,%s,%s" % (cell_widths, cell_widths, cell_depth)

        # randomise the zone axis (only up to 2)
        zone_h = 1#np.random.randint(0, 3)
        zone_k = 1#np.random.randint(0, 3)
        zone_l = 1#np.random.randint(0, 3)
        zone_string = "%s,%s,%s" % (zone_h, zone_k, zone_l)

        # random tilt perturbations (normal distribution)
        tilt_a = 0.#np.random.normal(scale=0.1)
        tilt_b = 0.#np.random.normal(scale=0.1)
        tilt_c = 0.#np.random.normal(scale=0.1)
        tilt_string = "%s,%s,%s" % (tilt_a, tilt_b, tilt_c)

        #
        # Do the simulation
        #

        # FNULL = open(os.devnull, 'w') # used to suppress the output
        subprocess.call([exe_filepath, cif_filepath, "-s"+cell_string, "-z"+zone_string, "-t"+tilt_string, "-o" + out_repeat_filepath, "-d"+device_string, "-c" + config_filepath])#, stdout=FNULL)


## Utility
def scale0to1(img):
    """Rescale image between 0 and 1"""

    img = img.astype(np.float32)

    min = np.min(img)
    max = np.max(img)

    if np.absolute(min-max) < 1.e-6:
        img.fill(0.5)
    else:
        img = (img-min) / (max-min)

    return img.astype(np.float32)

def disp(img):
    cv2.namedWindow('CV_Window', cv2.WINDOW_NORMAL)
    cv2.imshow('CV_Window', scale0to1(img))
    cv2.waitKey(0)
    return


def remap_phase(img):

    img = np.where(
        img < 0,
        2*img/np.pi + 1,
        1 - 2*img/np.pi
        )

    return img


if __name__ == "__main__":

    # with multiprocessing.Pool(num_gpus) as p:
    #     p.map(do_sim, cif_filepaths)

    filename = "20.cif"
    cif_filepath = os.path.join(CIFS_DIR, filename)

    do_sim(cif_filepath)

    amplitude_filepath = r"Z:\Jeffrey-Ede\models\wavefunctions\run_specific_output\0\0\EW_amplitude.tif"
    phase_filepath = r"Z:\Jeffrey-Ede\models\wavefunctions\run_specific_output\0\0\EW_phase.tif"

    amplitude = imread(amplitude_filepath, mode="F")
    phase = imread(phase_filepath, mode="F")

    disp(amplitude)
    disp(remap_phase(phase))

        